/*
 * Copyright 2008 Google Inc.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.google.gwt.jas.linker;

import com.google.gwt.core.ext.LinkerContext;
import com.google.gwt.core.ext.TreeLogger;
import com.google.gwt.core.ext.UnableToCompleteException;
import com.google.gwt.core.ext.linker.AbstractLinker;
import com.google.gwt.core.ext.linker.ArtifactSet;
import com.google.gwt.core.ext.linker.EmittedArtifact;
import com.google.gwt.core.ext.linker.LinkerOrder;
import com.google.gwt.core.ext.linker.LinkerOrder.Order;
import com.google.gwt.core.ext.linker.PublicResource;
import com.google.gwt.dev.util.Util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.util.Set;

/**
 * A Linker for producing a single JavaScript file from a GWT module. The use of
 * this Linker requires that the module has exactly one distinct compilation
 * result.
 */
@LinkerOrder(Order.POST)
public class FlexCompilerLinker extends AbstractLinker {

  public String getDescription() {
    return "FlexCompiler";
  }


  final static int TEMP_DIR_ATTEMPTS = 10000;

  public static File createTempDir() {
    File baseDir = new File(System.getProperty("java.io.tmpdir"));
    String baseName = System.currentTimeMillis() + "-";

    for (int counter = 0; counter < TEMP_DIR_ATTEMPTS; counter++) {
      File tempDir = new File(baseDir, baseName + counter);
      if (tempDir.mkdir()) {
        return tempDir;
      }
    }
    throw new IllegalStateException("Failed to create directory within "
        + TEMP_DIR_ATTEMPTS + " attempts (tried "
        + baseName + "0 to " + baseName + (TEMP_DIR_ATTEMPTS - 1) + ')');
  }

  public ArtifactSet link(TreeLogger logger, LinkerContext context,
      ArtifactSet artifacts)
      throws UnableToCompleteException {

      String flexSdkHome = System.getProperty("flexsdk.home", "/home/build/google3/third_party/flex4_sdk");
    logger.log(TreeLogger.INFO, "FlexCompiler home is " + flexSdkHome);
    String mxmlcCmd = flexSdkHome + "/bin/mxmlc -show-actionscript-warnings=false -default-frame-rate 60 -default-size 800 600";

    ArtifactSet toReturn = new ArtifactSet(artifacts);
    Set<PublicResource> resources = artifacts.find(PublicResource.class);
    File tempDir = createTempDir();

    for (PublicResource pub : resources) {
      if (pub.getPartialPath().endsWith(".as")) {
        try {
          logger.log(TreeLogger.INFO, "Writing " + pub.getPartialPath());
          File f = new File(tempDir, pub.getPartialPath());
          f.getParentFile().mkdirs();
          FileOutputStream fos = new FileOutputStream(f, false);
          Util.copy(logger, pub.getContents(logger), fos);
        } catch (FileNotFoundException e) {
          e.printStackTrace();
        }
      }
    }

    Set<EmittedArtifact> sources = artifacts.find(EmittedArtifact.class);
    for (EmittedArtifact e : sources) {

      if (e.getPartialPath().endsWith(".as") && !(e instanceof PublicResource)) {

        logger.log(TreeLogger.INFO, "" + e.getPartialPath());
        String asPath = e.getPartialPath();
        int lastSlash = asPath.lastIndexOf('/');
        if (lastSlash != -1) {
          asPath = asPath.substring(lastSlash + 1);
        }
        String swfPath = asPath.substring(0, asPath.length() - 2) + "swf";
        File swfFile = new File(tempDir, swfPath);
        String bootstrapPath =
            asPath.substring(0, asPath.length() - 2) + "nocache.js";
        try {
          File f = new File(tempDir, asPath);
          FileOutputStream fos = new FileOutputStream(f, false);
          Util.copy(logger, e.getContents(logger), fos);
          String absAsPath = new File(tempDir, asPath).getAbsolutePath();
          logger.log(TreeLogger.INFO, "Compiling " + mxmlcCmd + " " + absAsPath);
          Process proc = Runtime.getRuntime().exec(mxmlcCmd + " " + absAsPath);
          // crappy workaround for Process.waitFor() hanging issue
          Util.copy(logger, proc.getErrorStream(), System.err);
          proc.waitFor();
          toReturn.add(
              emitInputStream(logger, new FileInputStream(swfFile),  swfPath));
          toReturn.remove(e);
          String script = Util.readStreamAsString(
              getClass().getClassLoader().getResourceAsStream(
                  "com/google/gwt/jas/linker/swfobject.js"));
          script += "var so = new SWFObject('" + swfPath
              + "', 'movie', '800', '600', '60', '#ffffff', 'high');\n "
              + "so.addParam('quality', 'high');\n"
              + "so.addParam('wmode', 'opaque');\n " 
              + "so.addParam('allowScriptAccess','always');\n"
              + "so.write('flash_player');\n";
          toReturn.add(emitString(logger, script, bootstrapPath));

          // leave in on disk for now for now for debugging
          //		f.delete();
          //	        new File(swfPath).delete();
        } catch (Exception err) {
          logger.log(TreeLogger.ERROR, err.toString(), err);
        }
      }
    }

    return toReturn;
  }

}
